"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.FSRS = void 0;
const types_1 = require("./types");
const cardValidator_1 = require("./utils/cardValidator");
const timeFuncs_1 = require("./utils/timeFuncs");
class FSRS {
    constructor(parameters) {
        // Default FSRS parameters optimized for general use
        this.parameters = {
            requestRetention: 0.9, // Default 90%
            maximumInterval: 3650, // 10 years
            w: [
                0.4, 0.6, 2.4, 5.8, 4.93, 0.94, 0.86, 0.01, 1.49, 0.14, 0.94, 2.18, 0.05, 0.34, 1.26, 0.29,
                2.61, 0.0, 0.0,
            ],
            ...parameters,
        };
    }
    /**
     * Create a new card
     */
    createEmptyCard(now = new Date()) {
        return {
            due: new Date(now),
            stability: 0,
            difficulty: 0,
            elapsedDays: 0,
            scheduledDays: 0,
            reps: 0,
            lapses: 0,
            state: types_1.State.New,
            lastReview: undefined,
        };
    }
    schedule(card, now = new Date()) {
        if (!card)
            throw new Error("card cannot be null or undefined");
        if (card.lastReview && now < card.lastReview)
            throw new Error("Current time cannot be before the last review");
        return this.buildSchedulingCards(card, now);
    }
    /**
     * Schedule a card from raw data
     * Combines validation and scheduling in one operation
     * @param rawData - The raw card data
     * @param now - Current time (optional)
     * @returns Scheduling cards for all ratings
     * @throws Error if validation fails
     */
    scheduleRawCard(rawData, now = new Date()) {
        const card = this.convertRawCard(rawData);
        return this.schedule(card, now);
    }
    /**
     * Converts and validates raw card data from database/API
     * @param rawData - The raw card data
     * @returns A validated Card object
     * @throws Error if validation fails
     */
    convertRawCard(rawData) {
        return cardValidator_1.CardValidator.validateAndConvert(rawData);
    }
    /**
     * Batch convert and validate multiple raw cards
     * @param rawDataArray - Array of raw card data
     * @returns Object with valid cards and errors
     */
    convertRawCardBatch(rawDataArray) {
        return cardValidator_1.CardValidator.validateAndConvertBatch(rawDataArray);
    }
    buildSchedulingCards(card, now) {
        const cards = {};
        [types_1.Rating.Again, types_1.Rating.Hard, types_1.Rating.Good, types_1.Rating.Easy].forEach((rating) => {
            const scheduledCard = this.scheduleCard(card, rating, now);
            const reviewLog = this.buildReviewLog(card, scheduledCard, rating, now);
            const key = this.getRatingKey(rating);
            cards[key] = {
                card: scheduledCard,
                reviewLog,
            };
        });
        return cards;
    }
    scheduleCard(card, rating, now) {
        const newCard = { ...card };
        if (card.state === types_1.State.New) {
            const initDifficulty = this.initDifficulty(rating);
            const initStability = this.initStability(rating);
            newCard.difficulty = initDifficulty;
            newCard.stability = initStability;
            newCard.reps = 1;
            newCard.state = rating === types_1.Rating.Again ? types_1.State.Learning : types_1.State.Review;
            const interval = rating === types_1.Rating.Again ? 1 : Math.max(1, Math.round(initStability));
            newCard.scheduledDays = interval;
            newCard.due = this.addDays(now, interval);
        }
        else {
            const elapsedDays = card.lastReview ? (0, timeFuncs_1.calcElapsedDays)(card.lastReview, now) : 0;
            newCard.elapsedDays = elapsedDays;
            newCard.reps += 1;
            if (rating === types_1.Rating.Again) {
                newCard.lapses += 1;
                newCard.state = types_1.State.Relearning;
            }
            else {
                newCard.state = types_1.State.Review;
            }
            newCard.difficulty = this.nextDifficulty(card.difficulty, rating);
            newCard.stability = this.nextStability(card.difficulty, card.stability, elapsedDays, rating);
            const interval = this.nextInterval(newCard.stability);
            newCard.scheduledDays = interval;
            newCard.due = this.addDays(now, interval);
        }
        newCard.lastReview = new Date(now);
        return newCard;
    }
    buildReviewLog(card, scheduledCard, rating, now) {
        const elapsedDays = card.lastReview ? (0, timeFuncs_1.calcElapsedDays)(card.lastReview, now) : 0;
        return {
            rating,
            state: card.state,
            due: new Date(card.due),
            stability: card.stability,
            difficulty: card.difficulty,
            elapsedDays,
            lastElapsedDays: card.elapsedDays,
            scheduledDays: card.scheduledDays,
            review: new Date(now),
        };
    }
    getRatingKey(rating) {
        switch (rating) {
            case types_1.Rating.Again:
                return "again";
            case types_1.Rating.Hard:
                return "hard";
            case types_1.Rating.Good:
                return "good";
            case types_1.Rating.Easy:
                return "easy";
        }
    }
    // ----------------------------- FSRS Algorithm Core Functions -----------------------------
    initStability(rating) {
        return Math.max(this.parameters.w[rating - 1], 0.1);
    }
    initDifficulty(rating) {
        // Rotating
        return Math.min(Math.max(this.parameters.w[4] - (rating - 3) * this.parameters.w[5], 1), 10);
    }
    nextDifficulty(difficulty, rating) {
        // difficulty = how hard the card is 1 - 10
        // rating = 1-4 (Again, Hard, Good, Easy)
        const nextD = difficulty - this.parameters.w[6] * (rating - 3);
        // Prevents extremes on either scales low or high difficulty
        return Math.min(Math.max(this.meanReversion(this.parameters.w[4], nextD), 1), 10);
    }
    nextStability(difficulty, stability, elapsedDays, rating) {
        const hardPenalty = rating === types_1.Rating.Hard ? this.parameters.w[15] : 1;
        const easyBonus = rating === types_1.Rating.Easy ? this.parameters.w[16] : 1;
        if (rating === types_1.Rating.Again) {
            const baseStabilityForForgottenCards = this.parameters.w[11];
            return (baseStabilityForForgottenCards *
                Math.pow(difficulty, -this.parameters.w[12]) *
                (Math.pow(stability + 1, this.parameters.w[13]) - 1) *
                Math.exp(this.parameters.w[14] * (1 - this.retrievability(elapsedDays, stability))));
        }
        else {
            return (stability *
                (Math.exp(this.parameters.w[8]) *
                    (11 - difficulty) *
                    Math.pow(stability, -this.parameters.w[9]) *
                    (Math.exp(this.parameters.w[10] * (1 - this.retrievability(elapsedDays, stability))) -
                        1) *
                    hardPenalty *
                    easyBonus +
                    1));
        }
    }
    nextInterval(stability) {
        const interval = (stability / this.factor()) *
            (Math.pow(this.parameters.requestRetention, 1 / this.decay()) - 1);
        return Math.min(Math.max(Math.round(interval), 1), this.parameters.maximumInterval);
    }
    retrievability(elapsedDays, stability) {
        return Math.pow(1 + (this.factor() * elapsedDays) / stability, this.decay());
    }
    meanReversion(init, current) {
        return this.parameters.w[7] * init + (1 - this.parameters.w[7]) * current;
    }
    factor() {
        return Math.pow(0.9, 1 / this.decay()) - 1;
    }
    decay() {
        return -0.5;
    }
    addDays(date, days) {
        const result = new Date(date);
        result.setDate(result.getDate() + days);
        return result;
    }
    getRetrievability(card, now = new Date()) {
        if (card.state === types_1.State.New || !card.lastReview)
            return 1;
        const elapsedDays = (0, timeFuncs_1.calcElapsedDays)(card.lastReview, now);
        return this.retrievability(elapsedDays, card.stability);
    }
    updateParameters(newParameters) {
        this.parameters = { ...this.parameters, ...newParameters };
    }
    getParameters() {
        return { ...this.parameters };
    }
}
exports.FSRS = FSRS;
//# sourceMappingURL=fsrs.js.map